from __future__ import annotations

import enum
from typing import Optional

from ..dte.constants import TipoDteEnum


@enum.unique
class RcvKind(enum.Enum):
    """
    The kind of RCV.

    Definitions:

    * Registro de Compras (RC):
      > [..] todas las operaciones de compras realizadas por un contribuyente
      > de acuerdo a los DTEs recepcionados por el SII, complementado con los
      > documentos tributarios de compras, en soporte distinto al electrónico,
      > en el cual deberá indicarse la naturaleza de las operaciones en
      > cuanto a la procedencia e identificación del crédito fiscal.

    * Registro de Ventas (RV):
      > [..] todas las operaciones de ventas realizadas por un contribuyente
      > de acuerdo a los DTEs recibidos en el SII.

    Definitions source: RCV_FAQ_

    .. _RCV_FAQ: http://www.sii.cl/ayudas/ayudas_por_servicios/rcv_faqs.pdf

    """

    COMPRAS = 'COMPRAS'
    """RCV / compras."""

    VENTAS = 'VENTAS'
    """RCV / ventas."""

    def is_estado_contable_compatible(self, value: Optional[RcEstadoContable]) -> bool:
        if value is not None and not isinstance(value, RcEstadoContable):
            raise TypeError("Value must be None or a 'RcEstadoContable'.")

        if self == RcvKind.COMPRAS and value is not None:
            result = True
        elif self == RcvKind.VENTAS and value is None:
            result = True
        else:
            result = False

        return result


@enum.unique
class RcEstadoContable(enum.Enum):
    """
    The "Estado Contable" of a "Registro de compras" (RC).

    Applies to ``RcvKind.COMPRAS``.

    Definitions:

    * "Registro":
      > [..] los Documentos Tributarios Electrónicos (DTE) y no Electrónicos
      > que conforman la Información de Compras válida, la cual se utiliza para
      > la determinación impositiva y es considerada como el registro oficial
      > del Contribuyente y respaldo de su contabilidad.

    * "No incluir":
      > [..] los Documentos Tributarios Electrónicos (DTE) y no Electrónicos
      > que no deben incluirse en el Registro de Compras, en virtud de
      > corresponder a compras y pagos de servicios que no tienen relación con
      > las actividades económicas del contribuyente y por lo tanto no deben
      > afectar la determinación impositiva.

    * "Reclamado(s)":
      > [..] los Documentos Tributarios Electrónicos (DTE) que han sido
      > reclamados por el mismo receptor en el Registro de Aceptación o Reclamo
      > de un DTE.
      > Por encontrarse en estado "Reclamado", no es posible ingresar estos
      > documentos en el Registro de Compras vigente o considerarse para la
      > determinación impositiva.

    * "Pendiente(s)":
      > [..] los Documentos Tributarios Electrónicos (DTE) que han sido
      > recibidos en el SII, pero que se encuentran pendientes de otorgarse el
      > Acuse de Recibo o entregarse un Reclamo por parte del receptor [..].

    Definitions source: RCV_web_app_

    .. _RCV_web_app: https://www4.sii.cl/consdcvinternetui/#/index

    """

    REGISTRO = 'REGISTRO'
    NO_INCLUIR = 'NO_INCLUIR'
    RECLAMADO = 'RECLAMADO'
    PENDIENTE = 'PENDIENTE'


@enum.unique
class RcvTipoDocto(enum.IntEnum):

    """
    Enum of "Tipo de Documento" for the RCV domain.

    Unlike :class:`cl_sii.dte.constants.TipoDteEnum` this collection is not
    restricted to "documentos electrónicos". However, this is not a superset
    of the latter (e.g. "Guía electrónica de despacho" (52) is in
    ``TipoDteEnum`` but not in ``RcvTipoDocto``).

    Sources:

    * XML type (enum) ``DoctoType`` ("Tipos de Documentos") in
      official schema ``LibroCV_v10.xsd``.
      https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/LibroCV_v10.xsd#L1563-L1622

    * Values returned by SII endpoint
      https://www4.sii.cl/consdcvinternetui/services/data/facadeService/getDatosInicio

    * Constant ``cl_sii_api.rcv.AVAILABLE_DOCUMENT_TYPES`` in package
      ``cl-sii-api`` v0.2.2.
      https://github.com/fynpal/lib-cl-sii-api-python/blob/81b4a43/cl_sii_api/rcv.py

    """

    ###########################################################################
    # "facturas"
    ###########################################################################

    FACTURA_INICIO = 29
    """Factura de inicio."""

    FACTURA = 30
    """Factura."""

    FACTURA_ELECTRONICA = 33
    """Factura electrónica de venta."""

    FACTURA_NO_AFECTA_O_EXENTA = 32
    """Factura de venta, no afecta o exenta de IVA."""
    # aka 'Factura no Afecta o Exenta'
    # aka 'Factura de Venta de Bienes y Servicios No afectos o Exento de IVA'

    FACTURA_NO_AFECTA_O_EXENTA_ELECTRONICA = 34
    """Factura electrónica de venta, no afecta o exenta de IVA."""
    # aka 'Factura no Afecta o Exenta Electrónica'
    # aka 'Factura Electrónica de Venta de Bienes y Servicios No afectos o Exento de IVA'

    FACTURA_COMPRA = 45
    """Factura de compra."""

    FACTURA_COMPRA_ELECTRONICA = 46
    """Factura electrónica de compra."""
    # aka 'Factura de Compra Electrónica'
    # Name should have been 'Factura Electrónica de Compra'.

    FACTURA_EXPORTACION = 101
    """Factura de Exportación."""

    FACTURA_EXPORTACION_ELECTRONICA = 110
    """Factura Electrónica de Exportación."""
    # aka 'Factura de Exportación Electrónica'
    # Name should have been 'Factura Electrónica de Exportación'.

    ###########################################################################
    # "notas"
    ###########################################################################

    NOTA_DEBITO = 55
    """Nota de débito."""

    NOTA_DEBITO_ELECTRONICA = 56
    """Nota electrónica de débito."""
    # aka 'Nota de Débito Electrónica'

    NOTA_CREDITO = 60
    """Nota de crédito."""

    NOTA_CREDITO_ELECTRONICA = 61
    """Nota electrónica de crédito."""
    # aka 'Nota de Crédito Electrónica'

    NOTA_DEBITO_EXPORTACION = 104
    """Nota de débito de exportación."""

    NOTA_DEBITO_EXPORTACION_ELECTRONICA = 111
    """Nota electrónica de débito de exportación."""

    NOTA_CREDITO_EXPORTACION = 106
    """Nota de crédito de exportación."""

    NOTA_CREDITO_EXPORTACION_ELECTRONICA = 112
    """Nota electrónica de crédito de exportación."""

    ###########################################################################
    # "liquidación-factura"
    ###########################################################################

    # For more info about a "liquidación-factura" see
    #   'cl_sii.dte.constants.TipoDteEnum'.

    LIQUIDACION_FACTURA = 40
    """Liquidación-Factura."""

    LIQUIDACION_FACTURA_ELECTRONICA = 43
    """Liquidación-Factura Electrónica."""

    ###########################################################################
    # "Total Op. del mes Boleta X"
    ###########################################################################

    TOTAL_OP_DEL_MES_BOLETA_AFECTA = 35
    """Total Oper. del mes Boleta Afecta."""

    TOTAL_OP_DEL_MES_BOLETA_EXENTA = 38
    """Total Oper. del mes Boleta Exenta."""

    TOTAL_OP_DEL_MES_BOLETA_EXENTA_ELECTR = 41
    """Total Op. del mes Boleta Exenta Electr."""

    TOTAL_OP_DEL_MES_BOLETA_ELECTR = 39
    """Total Oper. del mes Boleta Electr."""

    ###########################################################################
    # uncommon
    ###########################################################################

    TIPO_47 = 47
    """Total del mes Vale Electrónico Especial."""

    TIPO_48 = 48
    """Total mes Comprobantes Pago Electrónico."""

    TIPO_102 = 102
    """Factura vta. Exenta a Zona Franca Prim."""

    TIPO_103 = 103
    """Liquidación."""

    TIPO_105 = 105
    """Total Op. mes Boleta Liq. Res. 1423/76."""

    TIPO_108 = 108
    """SRF Solicitud Registro de Factura."""

    TIPO_109 = 109
    """Factura Turista."""

    TIPO_901 = 901
    """Fact. Vta. Emp. Terr. Pref. Res. 1057/85."""

    TIPO_902 = 902
    """Conocimiento Embarque Marítimo o Aéreo."""

    TIPO_903 = 903
    """Documento Único de Salida (DUS)."""

    TIPO_904 = 904
    """Factura de Traspaso."""

    TIPO_905 = 905
    """Factura de Reexpedición."""

    TIPO_906 = 906
    """Total Op. del mes Boleta Vta. Módulo ZF."""

    TIPO_907 = 907
    """Facturas Venta Módulo ZF."""

    TIPO_909 = 909
    """Facturas Venta Módulo ZF."""

    TIPO_910 = 910
    """Solicitud Traslado Zona Franca (ZF)."""

    TIPO_911 = 911
    """Decl. de Ingreso a Zona Franca Primaria."""

    TIPO_914 = 914
    """Declaración de Ingreso (DIN)."""

    TIPO_919 = 919
    """Resumen Vtas. Pasajes Nac. sin Factura."""

    TIPO_920 = 920
    """Otros registros no Docum. Aumenta Débito."""

    TIPO_922 = 922
    """Otros registros no Doc. Disminuye Débito."""

    TIPO_924 = 924
    """Resumen Vtas. Pasajes Inter. sin Fact."""

    def as_tipo_dte(self) -> TipoDteEnum:
        """
        Return equivalent "Tipo DTE".

        :raises ValueError: if there is no equivalent one

        """
        try:
            value = TipoDteEnum(self.value)
        except ValueError as exc:
            raise ValueError(
                f"There is no equivalent 'TipoDteEnum' for 'RcvTipoDocto.{self.name}'.") from exc

        return value



"""
DTE-related constants.

Sources: official XML schemas 'SiiTypes_v10.xsd' and 'DTE_v10.xsd'.
https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/

"""
import enum


###############################################################################
# DTE fields / Folio
###############################################################################

# XML element 'DTEDefType/Documento/Encabezado/IdDoc/Folio'
# - description: "Folio del Documento Electronico"
# - XML type: 'FolioType'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/DTE_v10.xsd#L52-L56
# XML type 'FolioType' in official schema 'SiiTypes_v10.xsd'.
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/SiiTypes_v10.xsd#L153-L160

DTE_FOLIO_FIELD_TYPE = int
"""DTE field 'Folio' type."""
DTE_FOLIO_FIELD_MIN_VALUE = 1
"""DTE field 'Folio' min value."""
DTE_FOLIO_FIELD_MAX_VALUE = 10 ** 10
"""DTE field 'Folio' max value."""


###############################################################################
# DTE fields / Monto Total
###############################################################################

# XML element 'DTEDefType/Documento/Encabezado/Totales/MntTotal'
# - description: "Monto Total del DTE"
# - XML type: 'MontoType'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/DTE_v10.xsd#L1160-L1164
# XML type 'MontoType' in official schema 'SiiTypes_v10.xsd'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/SiiTypes_v10.xsd#L563-L570
# Specification for field 'Monto Total'
# - warning: In certain cases, such as whether negative values are allowed, the SII's specification
#   document may contradict the XML schema.
# - content:
#   > Campo: Monto Total <MntTotal>
#   > Descripción: [...]
#   > Largo máximo: 18
#   > Validación:
#   >   Valor Numérico de acuerdo a descripción.
#   >   En Liquidaciones-Factura, puede tomar valor negativo.
#   >   En Documentos de exportación es “0” (cero) si forma de pago es = 21 (sin pago).
#   > [...]
# - source: SII Chile. 2019-07-10. Formato Documentos Tributarios Electrónicos v2.2.
#   Contenido de Facturas y Documentos Asociados, Detalle por Zona, Encabezado, item nº 124.
#   http://www.sii.cl/factura_electronica/factura_mercado/formato_dte.pdf

DTE_MONTO_TOTAL_FIELD_TYPE = int
"""DTE field 'Monto Total' type."""
DTE_MONTO_TOTAL_FIELD_MIN_VALUE = -10 ** 18
"""DTE field 'Monto Total' min value."""
DTE_MONTO_TOTAL_FIELD_MAX_VALUE = 10 ** 18
"""DTE field 'Monto Total' max value."""


###############################################################################
# DTE fields / Tipo de DTE
###############################################################################

# XML element 'DTEDefType/Documento/Encabezado/IdDoc/TipoDTE'
# - description: "Tipo de DTE"
# - XML type: 'DTEType'
# - source:
#   https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/DTE_v10.xsd#L47-L51

DTE_TIPO_DTE_FIELD_TYPE = int
"""DTE field 'Tipo de DTE' type."""
DTE_TIPO_DTE_FIELD_MIN_VALUE = 1
"""DTE field 'Tipo de DTE' min value."""
# DTE_TIPO_DTE_FIELD_MAX_VALUE = ?
# """DTE field 'Tipo de DTE' max value."""


@enum.unique
class TipoDteEnum(enum.IntEnum):

    """
    Enum of "Tipo de DTE".

    Source: from official schema ``SiiTypes_v10.xsd``, the XML types (enums)
    ``DOCType``, ``DocType``, ``DTEType`` and ``DTEFacturasType`` which are
    VERY similar.
    https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/SiiTypes_v10.xsd

    Notes:
    * Enums ``DocType`` and ``DTEType`` have exactly the same elements
      (although descriptions differ).
    * The elements of the following enums are strictly subgroups of enum ``DOCType``:
      * ``DocType`` and ``DTEType``: same elements.
      * ``DTEFacturasType``
      * ``LIQType``: "Tipos de Liquidaciones".
      * ``EXPType``: "Tipos de Facturas de Exportacion".

    """

    FACTURA_ELECTRONICA = 33
    """Factura electrónica de venta."""

    FACTURA_NO_AFECTA_O_EXENTA_ELECTRONICA = 34
    """Factura electrónica de venta, no afecta o exenta de IVA."""
    # aka 'Factura no Afecta o Exenta Electrónica'
    # aka 'Factura Electrónica de Venta de Bienes y Servicios No afectos o Exento de IVA'

    LIQUIDACION_FACTURA_ELECTRONICA = 43
    """Liquidación-Factura Electrónica."""
    # For more info about a "liquidación-factura [electrónica]" see:
    #   - SII / FAQ / "¿Qué son las Liquidaciones-Factura?"
    #     http://www.sii.cl/preguntas_frecuentes/catastro/001_012_0247.htm
    #   - SII / FAQ / "¿Para qué se utiliza la Liquidación Factura Electrónica?"
    #     http://www.sii.cl/preguntas_frecuentes/catastro/001_012_3689.htm
    #   - SII / FAQ / "¿Qué documentos tributarios deben emitir las partes involucradas en un
    #     contrato de consignación?"
    #     http://www.sii.cl/preguntas_frecuentes/catastro/001_012_2619.htm
    #   - SII / resoluciones / "Resolución Exenta SII N°108 del 24 de Octubre del 2005.
    #     Establece normas y procedimientos de operación referente a la emisión de
    #     liquidaciones-facturas electrónicas"
    #     http://www.sii.cl/documentos/resoluciones/2005/reso108.htm

    FACTURA_COMPRA_ELECTRONICA = 46
    """Factura electrónica de compra."""
    # aka 'Factura de Compra Electrónica'
    # Name should have been 'Factura Electrónica de Compra'.

    GUIA_DESPACHO_ELECTRONICA = 52
    """Guía electrónica de despacho."""
    # aka 'Guía de Despacho Electrónica'

    NOTA_DEBITO_ELECTRONICA = 56
    """Nota electrónica de débito."""
    # aka 'Nota de Débito Electrónica'

    NOTA_CREDITO_ELECTRONICA = 61
    """Nota electrónica de crédito."""
    # aka 'Nota de Crédito Electrónica'

    # TODO: add
    #   110 Factura de exportación electrónica
    #   111 Nota de débito de exportación electrónica
    #   112 Nota de crédito de exportación electrónica
    # https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/SiiTypes_v10.xsd#L58-L60
    # https://github.com/fyndata/lib-cl-sii-python/blob/f57a326/cl_sii/data/ref/factura_electronica/schemas-xml/SiiTypes_v10.xsd#L708-L717
    #   See 'cl_sii.rcv.constants.RcvTipoDocto'
    #   Should 'is_factura' be true for a "Factura de exportación electrónica" (110) ?

    @property
    def is_factura(self) -> bool:
        if self is TipoDteEnum.FACTURA_ELECTRONICA:
            result = True
        elif self is TipoDteEnum.FACTURA_NO_AFECTA_O_EXENTA_ELECTRONICA:
            result = True
        elif self is TipoDteEnum.FACTURA_COMPRA_ELECTRONICA:
            result = True
        elif self is TipoDteEnum.LIQUIDACION_FACTURA_ELECTRONICA:
            result = True
        else:
            result = False

        return result

    @property
    def is_factura_venta(self) -> bool:
        if self is TipoDteEnum.FACTURA_ELECTRONICA:
            result = True
        elif self is TipoDteEnum.FACTURA_NO_AFECTA_O_EXENTA_ELECTRONICA:
            result = True
        elif self is TipoDteEnum.LIQUIDACION_FACTURA_ELECTRONICA:
            result = True
        else:
            result = False

        return result

    @property
    def is_factura_compra(self) -> bool:
        if self is TipoDteEnum.FACTURA_COMPRA_ELECTRONICA:
            result = True
        else:
            result = False

        return result

    @property
    def is_nota(self) -> bool:
        if self is TipoDteEnum.NOTA_DEBITO_ELECTRONICA:
            result = True
        elif self is TipoDteEnum.NOTA_CREDITO_ELECTRONICA:
            result = True
        else:
            result = False

        return result

    @property
    def emisor_is_vendedor(self) -> bool:
        return self.is_factura_venta

    @property
    def receptor_is_vendedor(self) -> bool:
        return self.is_factura_compra



"""
CTE Form 29 Data Models
=======================
"""

from __future__ import annotations

import dataclasses
import logging
from dataclasses import field as dc_field
from datetime import date
from typing import (
    Any,
    ClassVar,
    Iterator,
    Mapping, MutableMapping,
    Optional,
    Set,
    Type, Tuple,
    Union,
)

from cl_sii.rut import Rut
from cl_sii.rcv.data_models import PeriodoTributario


logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class CteForm29NaturalKey:
    contribuyente_rut: Rut
    periodo_tributario: PeriodoTributario
    folio: int


@dataclasses.dataclass(frozen=True)
class CteForm29:
    contribuyente_rut: Rut
    periodo_tributario: PeriodoTributario
    folio: int

    apellido_paterno_o_razon_social: Optional[str] = dc_field(default=None, repr=False)
    apellido_materno: Optional[str] = dc_field(default=None, repr=False)
    nombres: Optional[str] = dc_field(default=None, repr=False)
    calle_direccion: Optional[str] = dc_field(default=None, repr=False)
    numero_direccion: Optional[str] = dc_field(default=None, repr=False)
    comuna_direccion: Optional[str] = dc_field(default=None, repr=False)
    telefono: Optional[str] = dc_field(default=None, repr=False)
    correo_electronico: Optional[str] = dc_field(default=None, repr=False)
    representante_legal_rut: Optional[Rut] = dc_field(default=None, repr=False)

    total_a_pagar_en_plazo_legal: Optional[int] = dc_field(default=None, repr=False)
    total_a_pagar_con_recargo: Optional[int] = dc_field(default=None, repr=False)

    pct_condonacion: Optional[int] = dc_field(default=None, repr=False)
    num_res_condonacion: Optional[str] = dc_field(default=None, repr=False)
    fecha_condonacion: Optional[date] = dc_field(default=None, repr=False)

    tipo_declaracion: Optional[str] = dc_field(default=None, repr=False)
    banco: Optional[str] = dc_field(default=None, repr=False)
    medio_pago: Optional[str] = dc_field(default=None, repr=False)
    fecha_presentacion: Optional[date] = dc_field(default=None, repr=False)

    extra: Mapping[int, object] = dc_field(default_factory=dict, repr=False)
    """
    Any SII Form 29 codes that do not have their own field go in `extra`.
    """

    _strict_codes: bool = dc_field(default=False, repr=False)
    """
    Consider unknown codes as invalid and reject them.

    The default is `False` because the SII Form 29 has a large number of codes and not all of them
    are known to this class. Also, the SII may add new codes at any time in the future.
    """

    CODE_FIELD_MAPPING: ClassVar[Mapping[int, Optional[str]]]
    """
    Map SII Form 29 numeric codes to their respective field names.

    If a numeric code is valid, but no field has been created for it, use `None` so that this class
    considers it as "known". Numeric codes that are not included here are considered unknown or
    invalid, even if they appear in the SII Form 29.
    """

    CODE_FIELD_MAPPING = {
        1: 'apellido_paterno_o_razon_social',  # "APELLIDO PATERNO O RAZÓN SOCIAL"
        2: 'apellido_materno',  # Apellido Materno
        3: 'contribuyente_rut',  # "N DE RUT"
        5: 'nombres',  # Nombres
        6: 'calle_direccion',  # "DIRECCION"
        7: 'folio',  # "FOLIO"
        8: 'comuna_direccion',  # "COMUNA"
        9: 'telefono',  # Teléfono
        15: 'periodo_tributario',  # "PERIODO TRIBUTARIO"
        20: None,  # Exportaciones | Monto Neto
        30: None,  # "PPM ART. 84, A) PERD. ART. 90"
        48: None,  # "RET. IMP. ÚNICO TRAB. ART. 74 N 1 LIR"
        55: 'correo_electronico',  # Correo Electrónico
        60: 'pct_condonacion',  # "PORCENTAJE CONDONACION TGR"
        62: None,  # "PPM NETO DET."
        77: None,  # "REMANENTE DE CRÉDITO FISC."
        89: None,  # "IMP. DETERM. IVA DETERM."
        91: 'total_a_pagar_en_plazo_legal',  # "TOTAL A PAGAR DENTRO DEL PLAZO"
        92: None,  # "REAJUSTES"
        93: None,  # "Intereses y multas"
        94: 'total_a_pagar_con_recargo',  # "Total a pagar con recargo"
        110: None,  # Boletas | Cantidad
        111: None,  # Boletas | Débitos
        115: None,  # "TASA PPM 1ra. CATEGORIA"
        142: None,  # "VENTAS Y/O SERV. EXENTOS O NO GRAVADOS"
        151: None,  # "RET, TASAS DE 10 % SOBRE LAS RENT."
        153: None,  # "RET, TASAS DE 10% o 20% SOBRE LAS RENT."
        314: 'representante_legal_rut',  # Rut Representante Legal
        315: 'fecha_presentacion',  # "FECHA TIMBRE CAJA"
        500: None,  # "CANTIDAD FACTURAS"
        501: None,  # "LIQUIDACION DE FACTURAS"
        502: None,  # "DÉBITOS FACTURAS EMITIDAS"
        503: None,  # "CANTIDAD FACTURAS EMITIDAS"
        504: None,  # "REMANENTE CREDITO MES ANTERIOR"
        509: None,  # "CANT. DCTOS. NOTAS DE CRÉDITOS EMITIDAS"
        510: None,  # "DÉBITOS  NOTAS DE CRÉDITOS EMITIDAS"
        511: None,  # "CRÉD. IVA POR DCTOS. ELECTRONICOS"
        512: None,  # "CANT. DE DCTOS. NOTAS DE DÉBITO EMIT."
        513: None,  # "NOTAS DE DÉBITOS EMITIDAS"
        514: None,  # IVA por documentos electrónicos recibidos | Sin derecho a Crédito
        515: None,  # Facturas de Compra recibidas c/ret. total y Fact. de Inicio emitida | Cantidad
        516: None,  # Facturas de Compra recibidas con retención parcial | Cantidad
        517: None,  # Facturas de Compra recibidas con retención parcial | Débitos
        519: None,  # "CANT. DE DCTOS. FACT. RECIB. DEL GIRO"
        520: None,  # "CRÉDITO REC. Y REINT./FACT. DEL GIRO"
        521: None,  # "MONTO NETO / INTERNAS AFECTAS"
        524: None,  # "CANT. FACT. ACTIVO FIJO"
        525: None,  # "CRÉD.  RECUP. Y REINT. FACT. ACTIVO FIJO"
        527: None,  # "CANT. NOTAS DE CRÉDITO RECIBIDAS"
        528: None,  # "CRÉDITO RECUP. Y REINT NOTAS DE CRÉD"
        531: None,  # "CANT. NOTAS DE DÉBITO RECIBIDAS"
        532: None,  # "NOTAS DE DÉBITO CRÉD, RECUP. Y REINT."
        534: None,  # Declaraciones de Ingreso (DIN) importaciones del giro | Cantidad
        535: None,  # Declaraciones de Ingreso (DIN) importaciones del giro | Créd., Recup. y Reint.
        536: None,  # Declaraciones de Ingreso (DIN) import. activo fijo | Cantidad
        537: None,  # "TOTAL CRÉDITOS"
        538: None,  # "TOTAL DÉBITOS"
        547: None,  # "TOTAL DETERMINADO"
        553: None,  # Declaraciones de Ingreso (DIN) import. activo fijo | Créd., Recup. y Reint.
        562: None,  # "MONTO SIN DER. A CRED. FISCAL"
        563: None,  # "BASE IMPONIBLE"
        564: None,  # "CANT. DOC. SIN DER. A CRED. FISCAL"
        584: None,  # "CANT.INT.EX.NO GRAV.SIN DER. CRED.FISCAL"
        585: None,  # Exportaciones | Cantidad
        586: None,  # "CANT. VTAS. Y/O SERV. PREST. INT. EXENT."
        587: None,  # Facturas de Compra recibidas c/ret. total y Fact. de Inicio emitida | Monto
        595: None,  # "SUB TOTAL IMP. DETERMINADO ANVERSO"
        601: None,  # Fax
        610: 'numero_direccion',  # Nº Dirección
        708: None,  # "CANT. NOTAS CRED. EMIT. VALES MAQ. IVA"
        709: None,  # "MONTO NOTAS CRED. EMIT. VALES MAQ. IVA."
        755: None,  # IVA Postergado
        756: None,  # Casillero (checkbox) "Postergación pago de IVA"
        761: None,  # "CANT. FACT. SUPERMERCADOS Y SIMILARES"
        762: None,  # "CRÉD. FACT. SUPERMERCADOS Y SIMILARES"
        763: None,  # "CANT. FACT. POR VENTA BIENES INMUEBLES"
        764: None,  # "DÉB. FACT. POR VENTA BIENES INMUEBLES"
        765: None,  # "CANT. FACT. ADQ. O CONSTR. B. INMUEBLES"
        766: None,  # "DÉB. FACT. ADQ. O CONSTR. B. INMUEBLES"
        795: None,  # "MONTO DE CONDONACION SII"
        915: 'fecha_condonacion',  # "Fecha de Vigencia de Condonacion"
        922: 'num_res_condonacion',  # "NUMERO RESOLUCION SII AUTO. CONDONACION"
    }

    def __post_init__(self) -> None:
        # -----Set Fields from Extra-----

        new_extra: MutableMapping[int, object] = {}

        for code, value in self.extra.items():
            field_name = self.get_field_name(code, strict=self._strict_codes)

            if field_name is None:
                # If there's no field for the code; leave it in `extra`.

                new_extra[code] = value
            else:
                # There's a field for the code; remove it from `extra`.

                current_field_value = getattr(self, field_name)
                if current_field_value is None:
                    # The current field's value is empty, so we set it.
                    object.__setattr__(self, field_name, value)
                else:
                    # The current field's value is not empty and we do not overwrite it. We give
                    # precedence to the current field's value because it may have been previously
                    # converted to a different data type (e.g. `periodo_tributario` has a code, but
                    # the code's value must be parsed and converted to an instance of
                    # `PeriodoTributario`).
                    pass

        object.__setattr__(self, 'extra', new_extra)

        # -----Validations-----

        validate_field_type(self, 'contribuyente_rut', Rut)
        validate_field_type(self, 'periodo_tributario', PeriodoTributario)
        validate_field_type(self, 'folio', int)
        validate_field_type(self, 'representante_legal_rut', (Rut, type(None)))
        validate_field_type(self, 'fecha_presentacion', (date, type(None)))

        if not all(isinstance(code, int) for code in self.extra):
            raise TypeError("All codes in 'extra' must be integers")

        # TODO: Validate the type of the other fields.

        # -----Warn About Unknown Codes-----

        if not self._strict_codes:
            unknown_codes = self.get_all_codes(strict=False) - self.get_all_codes(strict=True)
            if unknown_codes:
                logger.warning(
                    "%s(contribuyente_rut=%r, periodo_tributario=%r, folio=%r)"
                    " contains invalid or unknown SII Form 29 codes: %s.",
                    self.__class__.__name__,
                    self.contribuyente_rut, self.periodo_tributario, self.folio,
                    ', '.join(str(code) for code in sorted(unknown_codes)),
                )

    @classmethod
    def get_field_name(cls, code: int, strict: bool = True) -> Optional[str]:
        """
        Return the field name for the SII Form 29 code if a field name has been defined for the
        code. Return ``None`` if the code is known, but no field name is associated with it.

        :param code: SII Form 29 code.
        :param strict: Whether to consider unknown codes as invalid and raise an exception.

        :raises KeyError: If ``code`` is invalid and ``strict`` is ``True``.
        """
        if not isinstance(code, int):
            raise TypeError(f"An integer is required (got type '{code.__class__.__name__}')")

        try:
            return cls.CODE_FIELD_MAPPING[code]
        except KeyError as e:
            if strict:
                raise KeyError(f"Invalid or unknown SII Form 29 code: {code}") from e
            else:
                return None

    @property
    def natural_key(self) -> CteForm29NaturalKey:
        return CteForm29NaturalKey(
            contribuyente_rut=self.contribuyente_rut,
            periodo_tributario=self.periodo_tributario,
            folio=self.folio,
        )

    def get_all_codes(self, strict: Optional[bool] = None) -> Set[int]:
        """
        Return a set with all codes.

        :param strict: Whether to include unknown codes.
        """
        strict = self._strict_codes if strict is None else strict

        if strict:
            return set(self.CODE_FIELD_MAPPING)
        else:
            return {*self.CODE_FIELD_MAPPING, *self.extra}

    def as_codes_dict(
        self,
        include_none: bool = True,
        strict: Optional[bool] = None,
    ) -> Mapping[int, object]:
        """
        Return a dictionary of SII Form 29 codes and values. Fields that do not have a code are not
        included.

        :param include_none: Include codes that have an empty value (i.e. ``None``).
        :param strict: Whether to include unknown codes.
        """
        strict = self._strict_codes if strict is None else strict

        obj_dict = {code: self[code] for code in self.get_all_codes(strict=strict)}

        if not include_none:
            obj_dict = {code: value for code, value in obj_dict.items() if value is not None}

        return obj_dict

    def __getitem__(self, key: int) -> Any:
        """
        Return the value of the SII Form 29 code ``key``.
        """
        field_name = self.get_field_name(key, strict=self._strict_codes)

        if field_name is None:
            # Field is valid, but no field name has been associated with it.
            return self.extra.get(key)
        else:
            # Field is valid and it has a field name.
            return getattr(self, field_name)

    def __iter__(self) -> Iterator:
        for dc_field_obj in dataclasses.fields(self):
            if not dc_field_obj.name.startswith('_'):  # Exclude private fields.
                yield (dc_field_obj.name, getattr(self, dc_field_obj.name))


def validate_field_type(
    obj: object,
    field_name: str,
    valid_types: Union[Type[Any], Tuple[Type[Any], ...]],
) -> None:
    if isinstance(valid_types, type):
        valid_types = (valid_types,)

    value = getattr(obj, field_name)
    if not isinstance(value, valid_types):
        raise TypeError(
            f"'{field_name}' must be an instance of"
            f" {', '.join(t.__name__ for t in valid_types)},"
            f" not '{value.__class__.__name__}'"
        )



import json
from pathlib import Path
from typing import Mapping


###############################################################################
# exceptions
###############################################################################

class JsonSchemaValidationError(Exception):
    """
    JSON data failed validation against a schema.

    Note: For the purpose of using this exception, anything that performs
        schema-like validations against JSON data is considered a 'schema', even
        if it is not a standard 'JSON Schema'.
    """


###############################################################################
# functions
###############################################################################

def read_json_schema(file_path: Path) -> Mapping[str, object]:
    """
    Instantiate an JSON schema object from a file.

    .. warning:: It is assumed that the schema is valid, and providing an
        invalid schema can lead to undefined behavior.

    :raises FileNotFoundError: If there is no file at ``file_path``.
    """
    with file_path.open(mode='rb') as file:
        content = json.load(file)

    if isinstance(content, Mapping):
        return content
    else:
        raise TypeError(
            f"Expected JSON file content to be a 'Mapping', not a '{content.__class__.__name__}'.",
        )
